#ifndef FILE_CAM_OFFSET
#define FILE_CAM_OFFSET
/*
    This header file is part of CadCam.

    This header file is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This header file is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CadCam.  If not, see <https://www.gnu.org/licenses/>.

    AUTOR     : Grotius
    DATE      : 14.06.2020
    REVISION  : --
    TODO      : --
    USAGE     : --
    PROTOCOL  : --
    NOTES     : 2d vector iterator workout:
                cam.erase(it); //erase complet contour
                cam.erase(it)->at(0); //erase complete contour
                cam[8].erase(cam[8].begin() + 0); //erase one object inside contour

                std::vector<std::vector<OBJECT>>::iterator it=cam.begin()+8;
                it->erase(it->begin()+0); //erase one object inside contour

                instert new object at a certain 2d vector place
                OBJECT obj;
                obj.type="line";
                obj.red=255; obj.alpha=255;
                obj.end.x=100;
                cam[8].insert(cam[8].begin()+1,obj);
*/
#include <iostream>
#include <algorithm>
#include <structs.h>
#include <math.h>
#include <lib/lib_intersect.h>
#include <lib/lib_offset.h>
#include <lib/lib_split.h>
#include <lib/lib_pocket.h>
#include <lib/lib_connect.h>

std::vector<OBJECT> find_intersections(std::vector<OBJECT> array);
OBJECT construct_arc(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1, double d);
std::vector<OBJECT> connect_objects(std::vector<OBJECT> array);
OBJECT raw_offset(std::vector<OBJECT>::iterator it);

OBJECT raw_offset(std::vector<OBJECT>::iterator it){
    if(it->type=="line"){
        OBJECT obj=line_offset(it,offset);
        obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
        return obj;
    }
    if(it->type=="arc"){
        if(std::round(it->red)==255 && std::round(it->green)==255){ //yellow, inside offset
            OBJECT obj=arc_offset(*it,offset);
            obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
            return obj;
        }
        if(std::round(it->red)==255 && std::round(it->green)==0){ //red, outside offset
            OBJECT obj=arc_offset(*it,abs(offset));
            obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
            return obj;
        }
    }
    if(it->type=="circle"){
        bool checked=true;
        double temp_offset=offset;
        if(temp_offset<0){
            if(abs(temp_offset)>it->radius){ //dont add circles with negative radius
                checked=false;
            }
        }
        if(checked){cam.clear();
            if(std::round(it->red)==255 && std::round(it->green)==255){ //yellow, inside offset
                OBJECT obj;
                obj=circle_offset(it,offset);
                obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
                return obj;
            }
        }
        if(std::round(it->red)==255 && std::round(it->green)==0){ //red, outside offset
            OBJECT obj;
            obj=circle_offset(it,abs(offset));
            obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
            return obj;
        }
    }
    if(it->type=="linestrip" || it->type=="rectangular" || it->type=="polygon"){
        OBJECT obj;
        obj=linestrip_offset(it,offset);
        obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
        return obj;
    }
    if(it->type=="spline"){
        OBJECT obj;
        obj=spline_offset_as_linestrip(it,resolution,offset);
        obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
        return obj;
    }
    if(it->type=="ellipse"){
        OBJECT obj;
        obj=ellipse_offset_as_linestrip(it,resolution,offset);
        obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
        return obj;
    }
    if(it->type=="bezier"){
        OBJECT obj;
        obj=bezier_offset_as_linestrip(it,resolution,offset);
        obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
        return obj;
    }

}

void cam_offset(){

    //test.clear();
    std::vector<CONTOUR>::iterator it;
    std::vector<int>::iterator it1;
    std::vector<OBJECT>::iterator it2;


    //copy first offset
    for(it=contours.begin(); it<contours.end(); it++){
        if(it->area<0 ||it->depth>0){ //we got a yellow contour, perfect for testing a pocket algoritme
            //std::cout<<"yellow contour found"<<std::endl;
            std::vector<OBJECT> a,b,c,d,e,f;

            for(it1=it->sequence.begin(); it1<it->sequence.end(); it1++){
                int object = *it1;
                //std::cout<<"int object: "<<object<<std::endl;
                it2=cad.begin()+object;

                OBJECT obj=raw_offset(it2);
                a.push_back(obj);
            }
            b=connect_objects(a);
            c=find_intersections(b);
            //e=split_primitives(d);
            //f=pocket_algoritme(e);
            cam.insert(cam.end(),c.begin(),c.end());

            //next_offset(f);
        }
    }
}

std::vector<OBJECT> connect_objects(std::vector<OBJECT> array){ //for inside offset

    std::vector<OBJECT> a,b;
    std::vector<OBJECT>::iterator it,it1;

    for(unsigned int i=0; i<array.size(); i++){ //connect lines if corners are inside
        it=array.begin();
        it1=array.begin()+1;

        if(it->type=="line" && it1->type=="line"){
            double xp = it1->end.x;
            double yp = it1->end.y;
            double x1 = it->start.x;
            double y1 = it->start.y;
            double x2 = it->end.x;
            double y2 = it->end.y;
            double d = ((xp-x1)*(y2-y1))-((yp-y1)*(x2-x1));
            if(d<0){
                line_line_connect(it,it1);
            }
        }
        if(it->type=="linestrip" && it1->type=="line"){
            double xp = it1->end.x;
            double yp = it1->end.y;
            double x1 = it->control.back().x;
            double y1 = it->control.back().y;
            double x2 = it->end.x;
            double y2 = it->end.y;
            double d = ((xp-x1)*(y2-y1))-((yp-y1)*(x2-x1));
            if(d<0){
                linestrip_line_connect(it,it1);
            }
        }
//        if(it->type=="line" && it1->type=="linestrip"){
//            double xp = it1->start.x;
//            double yp = it1->start.y;
//            double x1 = it->start.x;
//            double y1 = it->start.y;
//            double x2 = it->end.x;
//            double y2 = it->end.y;
//            double d = ((xp-x1)*(y2-y1))-((yp-y1)*(x2-x1));
//            if(d<0){
//                line_linestrip_connect(it,it1);
//            }
//        }




        if(it->type=="arc" && it1->type=="arc"){ //only connects if points are on the segments
            //arc_arc_connect(it,it1);
        }
        if(it->type=="line" && it1->type=="arc"){ //only connects if points are on the segments
            //line_arc_connect(it,it1);
        }
        if(it->type=="arc" && it1->type=="line"){ //only connects if points are on the segments
            //arc_line_connect(it,it1);
        }
        std::rotate(array.begin(),array.begin()+1,array.end());
    }

    for(unsigned int i=0; i<array.size(); i++){ //for line-line, place arc connectors if corners are outside
        it=array.begin();
        it1=array.begin()+1;

        a.push_back(*it);

        if(it->type=="line" && it1->type=="line"){
            double xp = it1->end.x;
            double yp = it1->end.y;
            double x1 = it->start.x;
            double y1 = it->start.y;
            double x2 = it->end.x;
            double y2 = it->end.y;
            double d = ((xp-x1)*(y2-y1))-((yp-y1)*(x2-x1));
            if(d>0){
                OBJECT arc=construct_arc(it,it1,d);
                a.push_back(arc);
            }
        }
        std::rotate(array.begin(),array.begin()+1,array.end());
    }

    array=a;
    a.clear();
    for(unsigned int i=0; i<array.size(); i++){ //for arc-line, line-arc, arc-arc, place arc connectors if they don't connect.
        it=array.begin();
        it1=array.begin()+1;

        a.push_back(*it);

        if((it->type=="arc" && it1->type=="line")||(it->type=="line" && it1->type=="arc")||(it->type=="arc" && it1->type=="arc")){

            //check if it endpoint is connected to it1 startpoint, if not, place g2 arc

            bool is_point=point_is_point(it->end,it1->start,0.01);
            if(!is_point){
                OBJECT arc=construct_arc(it,it1,1); //g2
                a.push_back(arc);
            }




        }
        std::rotate(array.begin(),array.begin()+1,array.end());
    }







    return a;




    //    std::vector<OBJECT> a,b;
    //    std::vector<OBJECT>::iterator it,it1;


    //    //chamfers
    //    for(it=array.begin(); it<array.end(); it++){
    //        it->sector=0; //reset sectors.
    //        if(it<array.end()-1){ //between cycle
    //            it1=it+1;
    //        }
    //        if(it==array.end()-1){ //close cycle
    //            it1=array.begin();
    //        }
    //        line_line_connect(it,it1);
    //    }



    //    //check if offset is closed, if not, place connectors
    //    for(it=array.begin(); it<array.end(); it++){
    //        if(it<array.end()-1){ //between cycle
    //            it1=it+1;
    //        }
    //        if(it==array.end()-1){ //close cycle
    //            it1=array.begin();
    //        }

    //        a.push_back(*it);

    //        if(it->end.x<it1->start.x+0.01 && it->end.x>it1->start.x-0.01 && it->end.y<it1->start.y+0.01 && it->end.y>it1->start.y-0.01){
    //            //ok
    //        } else {
    //            //drop connector
    //            OBJECT obj;
    //            obj=*it;
    //            obj.start.x=it->end.x;
    //            obj.start.y=it->end.y;
    //            obj.end.x=it1->start.x;
    //            obj.end.y=it1->start.y;
    //            a.push_back(obj);
    //        }
    //    }


    //    //check if offset is closed again
    //    for(it=array.begin(); it<array.end(); it++){
    //        if(it<array.end()-1){ //between cycle
    //            it1=it+1;
    //        }
    //        if(it==array.end()-1){ //close cycle
    //            it1=array.begin();
    //        }

    //        if(it->end.x<it1->start.x+0.01 && it->end.x>it1->start.x-0.01 && it->end.y<it1->start.y+0.01 && it->end.y>it1->start.y-0.01){

    //        } else { std::cout<<"object not closed"<<std::endl;}

    //    }

    //    //check line lenghts
    //    for(it=array.begin(); it<array.end(); it++){
    //        if(it<array.end()-1){ //between cycle
    //            it1=it+1;
    //        }
    //        if(it==array.end()-1){ //close cycle
    //            it1=array.begin();
    //        }

    //        double l=sqrt(pow(it->start.x-it->start.y,2)+pow(it->end.x-it->end.y,2));
    //        std::cout<<"line lenght: "<<l<<std::endl;
    //    }




}

std::vector<OBJECT> find_intersections(std::vector<OBJECT> array)
{
    std::vector<POINT> p;
    std::vector<OBJECT>::iterator it,it1;
    for(it=array.begin(); it<array.end(); it++){

        for(it1=array.begin(); it1<array.end(); it1++){
            if(it!=it1){

                if(it->type=="line"&&it1->type=="line"){
                    line_line_intersect(it,it1);
                }
                if(it->type=="arc"&&it1->type=="arc"){
                    arc_arc_intersect(it,it1);
                }
                if(it->type=="line"&&it1->type=="arc"){
                    line_arc_intersect(it,it1);
                }
                if(it->type=="arc"&&it1->type=="line"){
                    arc_line_intersect(it,it1);
                }
            }
        }
    }
    return array;
}

OBJECT construct_arc(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1, double d){

    OBJECT obj;
    obj.type="arc";
    obj.red=255; obj.green=0; obj.blue=255; obj.alpha=255;
    obj.start.x=it->end.x;
    obj.start.y=it->end.y;
    obj.end.x=it1->start.x;
    obj.end.y=it1->start.y;
    obj.radius=abs(offset);
    obj.control.resize(1);

    double radsq=obj.radius*obj.radius;
    double q=sqrt(((obj.end.x-obj.start.x)*(obj.end.x-obj.start.x))+((obj.end.y-obj.start.y)*(obj.end.y-obj.start.y)));
    double x3=(obj.start.x+obj.end.x)/2;
    double y3=(obj.start.y+obj.end.y)/2;
    double pi_angle_start=0,pi_angle_end=0;

    if(d>0){ //G2
        obj.center.x=x3-sqrt(radsq-((q/2)*(q/2)))*((obj.start.y-obj.end.y)/q);
        obj.center.y=y3-sqrt(radsq-((q/2)*(q/2)))*((obj.end.x-obj.start.x)/q);

        pi_angle_end=atan2(obj.start.y-obj.center.y, obj.start.x-obj.center.x);
        pi_angle_start=atan2(obj.end.y-obj.center.y, obj.end.x-obj.center.x);
        if(pi_angle_start>pi_angle_end){pi_angle_end+=2*M_PI;}
        obj.control.front().x=obj.center.x+(cos((pi_angle_start+pi_angle_end)/2)*obj.radius);
        obj.control.front().y=obj.center.y+(sin((pi_angle_start+pi_angle_end)/2)*obj.radius);

    }
    if(d<0){ //G3
        obj.center.x=x3+sqrt(radsq-((q/2)*(q/2)))*((obj.start.y-obj.end.y)/q);
        obj.center.y=y3+sqrt(radsq-((q/2)*(q/2)))*((obj.end.x-obj.start.x)/q);

        pi_angle_end=atan2(obj.end.y-obj.center.y, obj.end.x-obj.center.x);
        pi_angle_start=atan2(obj.start.y-obj.center.y, obj.start.x-obj.center.x);
        if(pi_angle_start>pi_angle_end){pi_angle_end+=2*M_PI;}
        obj.control.front().x=obj.center.x+(cos((pi_angle_start+pi_angle_end)/2)*obj.radius);
        obj.control.front().y=obj.center.y+(sin((pi_angle_start+pi_angle_end)/2)*obj.radius);
    }
    return obj;
}

#endif



//        if(it>array.begin()){
//             double l=sqrt(pow(it->start.x-it->start.y,2)+pow(it->end.x-it->end.y,2));
//             if(l>0){
//                a.push_back(*it);
//             }
//        }
//        if(connect){

//            OBJECT obj;
//            obj=*it;
//            obj.start.x=it->end.x;
//            obj.start.y=it->end.y;
//            obj.end.x=it1->start.x;
//            obj.end.y=it1->start.y;
//            a.push_back(obj);
//        }


//it=array.begin();
//a.push_back(*it);


